unit CSVOps;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil;

type
  ECSVTableException = class(Exception);
  PCSVCell = ^TCSVCell;

  { TCSVCell }

  TCSVCell = object
  private
    FValue: string;
  public
    property Value: string read FValue write FValue;
  end;

  { TCSVTable }

  TCSVTable = class
  private
    FHeight: Integer;
    FModified: Boolean;
    FWidth: Integer;
    FCells: array of array of TCSVCell;
    function GetCell(X, Y: Integer): PCSVCell;
    function GetCellString(X, Y: Integer): string;
    procedure SetCellString(X, Y: Integer; const AValue: string);
    procedure SetHeight(const AValue: Integer); inline;
    procedure SetWidth(const AValue: Integer); inline;
  public
    procedure Resize(AWidth, AHeight: Integer);
    function AddField(AField: string): PCSVCell;
    procedure RemoveField(AIndex: Integer);
    procedure SwapFields(AIndex1, AIndex2: Integer);
    procedure InsertRow(AIndex: Integer);
    procedure DeleteRow(AIndex: Integer);
    procedure SwapRows(AIndex1, AIndex2: Integer);
    procedure SaveToStream(AStream: TStream);
    procedure SaveToFile(AFileName: string);
    procedure LoadFromStream(AStream: TStream);
    procedure LoadFromFile(AFileName: string);
    property Modified: Boolean read FModified write FModified;
    property Cell[X, Y: Integer]: PCSVCell read GetCell;
    property CellString[X, Y: Integer]: string read GetCellString write SetCellString;
    property Width: Integer read FWidth write SetWidth;
    property Height: Integer read FHeight write SetHeight;
  end;

implementation

{ TCSVTable }

procedure TCSVTable.SetHeight(const AValue: Integer);
begin
  Resize(FWidth, AValue);
end;

function TCSVTable.GetCell(X, Y: Integer): PCSVCell;
begin
  if (X >= 0) and (Y >= 0) and (X < Width) and (Y < Height) then
    Result:=@FCells[Y, X]
  else
    if (Width=0) or (Height=0) then
      raise ECSVTableException.Create('The table is empty')
    else
      raise ECSVTableException.Create('Cell coordinates ' + IntToStr(X) + ',' + IntToStr(Y) + ' out of range 0,0..' + IntToStr(Width - 1) + ',' + IntToStr(Height - 1));
end;

function TCSVTable.GetCellString(X, Y: Integer): string;
begin
  if (X >= 0) and (Y >= 0) and (X < Width) and (Y < Height) then
    Result:=FCells[Y, X].FValue
  else
    if (Width=0) or (Height=0) then
      raise ECSVTableException.Create('The table is empty')
    else
      raise ECSVTableException.Create('Cell coordinates ' + IntToStr(X) + ',' + IntToStr(Y) + ' out of range 0,0..' + IntToStr(Width - 1) + ',' + IntToStr(Height - 1));
end;

procedure TCSVTable.SetCellString(X, Y: Integer; const AValue: string);
begin
  if (X >= 0) and (Y >= 0) and (X < Width) and (Y < Height) then begin
    if FCells[Y, X].FValue=AValue then Exit;
    FCells[Y, X].FValue:=AValue;
    Modified:=True;
  end else
    if (Width=0) or (Height=0) then
      raise ECSVTableException.Create('The table is empty')
    else
      raise ECSVTableException.Create('Cell coordinates ' + IntToStr(X) + ',' + IntToStr(Y) + ' out of range 0,0..' + IntToStr(Width - 1) + ',' + IntToStr(Height - 1));
end;

procedure TCSVTable.SetWidth(const AValue: Integer);
begin
  Resize(AValue, Height);
end;

procedure TCSVTable.Resize(AWidth, AHeight: Integer);
var
  i: Integer;
begin
  if (AWidth=FWidth) and (AHeight=FHeight) then Exit;
  if AHeight <> FHeight then begin
    SetLength(FCells, AHeight);
    if AHeight > 0 then SetLength(FCells[AHeight - 1], AWidth);
    FHeight:=AHeight;
  end;
  if AWidth <> FWidth then begin
    for i:=0 to FHeight - 1 do
      SetLength(FCells[i], AWidth);
    FWidth:=AWidth;
  end;
  Modified:=True;
end;

function TCSVTable.AddField(AField: string): PCSVCell;
begin
  if Height=0 then
    Resize(Width + 1, Height + 1)
  else
    Resize(Width + 1, Height);
  Result:=@FCells[0, Width - 1];
  Result^.Value:=AField;
end;

procedure TCSVTable.RemoveField(AIndex: Integer);
var
  x, y: Integer;
begin
  if (AIndex < 0) or (AIndex >= Width) then
    raise ECSVTableException.Create('Field index ' + IntToStr(AIndex) + ' is outside of the table field range');
  for x:=AIndex to Width - 2 do
    for y:=0 to Height - 1 do FCells[y, x]:=FCells[y, x + 1];
  if Width=1 then
    Resize(0, 0)
  else
    Resize(Width - 1, Height);
end;

procedure TCSVTable.SwapFields(AIndex1, AIndex2: Integer);
var
  y: Integer;
  Tmp: TCSVCell;
begin
  if (AIndex1 < 0) or (AIndex1 >= Width) then
    raise ECSVTableException.Create('First field index ' + IntToStr(AIndex1) + ' is outside of the table field range');
  if (AIndex2 < 0) or (AIndex2 >= Width) then
    raise ECSVTableException.Create('Second field index ' + IntToStr(AIndex2) + ' is outside of the table field range');
  for y:=0 to Height - 1 do begin
    Tmp:=FCells[y, AIndex1];
    FCells[y, AIndex1]:=FCells[y, AIndex2];
    FCells[y, AIndex2]:=Tmp;
  end;
  Modified:=True;
end;

procedure TCSVTable.InsertRow(AIndex: Integer);
var
  x, y: Integer;
begin
  Resize(Width, Height + 1);
  if AIndex >= Height then Exit;
  for y:=Height - 1 downto AIndex + 1 do
    for x:=0 to Width - 1 do FCells[y, x]:=FCells[y - 1, x];
  for x:=0 to Width - 1 do FCells[AIndex, x].Value:='';
end;

procedure TCSVTable.DeleteRow(AIndex: Integer);
var
  x, y: Integer;
begin
  if (AIndex < 0) or (AIndex >= Height) then
    raise ECSVTableException.Create('Row index ' + IntToStr(AIndex) + ' is outside of the table row range');
  for y:=AIndex to Height - 2 do
    for x:=0 to Width - 1 do FCells[y, x]:=FCells[y + 1, x];
  Resize(Width, Height - 1);
end;

procedure TCSVTable.SwapRows(AIndex1, AIndex2: Integer);
var
  x: Integer;
  Tmp: TCSVCell;
begin
  if (AIndex1 < 0) or (AIndex1 >= Height) then
    raise ECSVTableException.Create('First row index ' + IntToStr(AIndex1) + ' is outside of the table row range');
  if (AIndex2 < 0) or (AIndex2 >= Height) then
    raise ECSVTableException.Create('Second row index ' + IntToStr(AIndex2) + ' is outside of the table row range');
  for x:=0 to Width - 1 do begin
    Tmp:=FCells[AIndex1, x];
    FCells[AIndex1, x]:=FCells[AIndex2, x];
    FCells[AIndex2, x]:=Tmp;
  end;
  Modified:=True;
end;

procedure TCSVTable.SaveToStream(AStream: TStream);
var
  x, y: Integer;
  b: UTF8String;

  function EncodeCSVString(s: string): string;
  var
    i: Integer;
    Quote: Boolean = False;
  begin
    Result:='';
    for i:=1 to Length(s) do case s[i] of
      '"': begin
        Result:=Result + '""';
        Quote:=True;
      end;
      #10, #13, ',': begin
        Result:=Result + s[i];
        Quote:=True;
      end;
      else Result:=Result + s[i];
    end;
    if Quote then Result:='"' + Result + '"';
  end;

begin
  for y:=0 to Height - 1 do begin
    for x:=0 to Width - 1 do begin
      if x > 0 then AStream.WriteByte($2C);
      b:=EncodeCSVString(CellString[x, y]);
      AStream.Write(b[1], Length(b));
    end;
    AStream.WriteByte($0D);
    AStream.WriteByte($0A);
  end;
end;

procedure TCSVTable.SaveToFile(AFileName: string);
var
  Stream: TFileStream;
begin
  Stream:=TFileStream.Create(AFileName, fmCreate);
  SaveToStream(Stream);
  FreeAndNil(Stream);
end;

procedure TCSVTable.LoadFromStream(AStream: TStream);
var
  Code: UTF8String = '';
  Head, Len, Col, Row: Integer;
  ChangeRow: Boolean;
  CellValue: string;

  function ParseQuottedCell: string;
  begin
    Result:='';
    Inc(Head); // skip doublequote
    while Head <= Len do begin
      if Code[Head]='"' then begin
        Inc(Head);
        if (Head > Len) or (Code[Head] <> '"') then Break;
      end;
      Result:=Result + Code[Head];
      Inc(Head);
    end;
    while (Head <= Len) and (not (Code[Head] in [',', #10, #13])) do Inc(Head);
  end;

  function ParseNudeCell: string;
  begin
    Result:='';
    while Head <= Len do begin
      if Code[Head] in [',', #10, #13] then Break;
      Result:=Result + Code[Head];
      Inc(Head);
    end;
  end;

  function ParseCell: string;
  begin
    if Code[Head]='"' then Result:=ParseQuottedCell else Result:=ParseNudeCell;
  end;

begin
  Resize(0, 0);
  while AStream.Position < AStream.Size do Code:=Code + AnsiChar(AStream.ReadByte);
  Head:=1;
  Len:=Length(Code);
  Col:=0;
  Row:=0;
  while Head <= Len do begin
    CellValue:=ParseCell;
    ChangeRow:=False;
    if (Head <= Len) and (Code[Head] in [',', #10, #13]) then begin
      if Code[Head]=',' then
        Inc(Head)
      else begin
        Inc(Head);
        if (Head <= Len) and (Code[Head - 1]=#13) and (Code[Head]=#10) then Inc(Head);
        ChangeRow:=True;
      end;
    end;
    if Width <= Col then Width:=Col + 1;
    if Height <= Row then Height:=Row + 1;
    CellString[Col, Row]:=CellValue;
    if ChangeRow then begin
      Col:=0;
      Inc(Row);
    end else Inc(Col);
  end;
  Modified:=False;
end;

procedure TCSVTable.LoadFromFile(AFileName: string);
var
  Stream: TFileStream;
begin
  Stream:=TFileStream.Create(AFileName, fmOpenRead);
  LoadFromStream(Stream);
  FreeAndNil(Stream);
end;

end.

